@@ -2,14 +2,16 @@
 
 import asyncio
 import logging
+from collections import defaultdict
 from datetime import datetime
 from functools import partial, lru_cache
 from itertools import chain
 from operator import itemgetter
 from typing import MutableMapping, List, Union, Coroutine, Type, Sequence, Dict, Callable, Awaitable, Optional, \
-    Iterator, NoReturn
+    Iterator, NoReturn, Tuple
 
-from patchwork.core import Task, AsyncSubscriber
+import pytz as pytz
+from patchwork.core import Task, AsyncSubscriber, TaskMetadata
 from patchwork.core.config.base import ComponentConfig
 from patchwork.core.typing import FuncPath
 from patchwork.core.utils import cached_property, Flag
@@ -34,13 +36,13 @@ ExceptionHandler = Callable[[Task, Exception], Awaitable]
 
 
 async def check_task_middleware(task: Task, receiver: AsyncSubscriber, finalizer: FinalizerType):
-    now = datetime.utcnow().timestamp()
-    if task.meta.HasField('expires') and now > task.meta.expires.ToSeconds():
+    now = datetime.now(pytz.UTC)
+    if task.meta.expires and task.meta.expires < now:
         # task has expired
         raise TaskFatal(reason="Task has expired")
-    elif task.meta.HasField('not_before') and now < task.meta.not_before.ToSeconds():
+    elif task.meta.not_before and task.meta.not_before > now:
         # task is not ready yet
-        raise TaskRetry(countdown=(now - task.meta.not_before))
+        raise TaskRetry(countdown=(now - task.meta.not_before).total_seconds())
 
 
 async def default_task_exception_handler(task: Task, exception: Exception) -> NoReturn:
@@ -78,6 +80,10 @@ class BaseExecutor(Module):
 
         # executor termination timeout in seconds
         terminate_timeout: int = 10
+
+        # retry limit for the worker, it will be forced even if task max retries is greater
+        max_retries: int = 10
+
         unit: ComponentConfig[Type[ProcessingUnit]] = ComponentConfig(InlineUnit)
         middlewares: Sequence[FuncPath] = [
             FuncPath(check_task_middleware)
@@ -86,7 +92,7 @@ class BaseExecutor(Module):
     EVENTS = Module.EVENTS + ('task', )
 
     _loop_task: asyncio.Future
-    _routers: List[TaskRouter]
+    _routers: List[Tuple[str, TaskRouter]]
     _unit_monitor_task: asyncio.Future
     _busy: Flag
 
@@ -125,14 +131,25 @@ class BaseExecutor(Module):
     def unit(self) -> ProcessingUnit:
         return self.settings.unit.instantiate(parent=self)
 
-    def add_router(self, router: TaskRouter):
+    def add_router(self, router: TaskRouter, prefix: str = ''):
         """
         Adds a router to the executor. Routers are responsible for routing tasks to proper handlers.
+        :param prefix:
         :param router:
         :return:
         """
         router.bind(self.worker)
-        self._routers.append(router)
+        for idx, rdata in enumerate(self._routers):
+            if prefix == rdata[0]:
+                raise ValueError(f"routers prefix conflict: prefix is already taken by another router")
+
+            if prefix.startswith(rdata[0]):
+                # if given prefix starts with one of already registered prefix, put
+                # new router on higier priority then existing one
+                self._routers.insert(idx, (prefix, router))
+                break
+        else:
+            self._routers.append((prefix, router))
 
     def add_middleware(self, middleware: MiddlewareType):
         """
@@ -329,13 +346,21 @@ class BaseExecutor(Module):
         :param bump_attempts:
         :return:
         """
-        new_task = Task()
-        new_task.CopyFrom(task)
-        new_task.uuid = ''
+        new_task = Task(
+            task_type=task.task_type,
+            correlation_id=task.correlation_id,
+            payload=task.payload
+        )
+        new_task.meta = TaskMetadata(
+            **task.meta.dict()
+        )
 
         if bump_attempts:
             new_task.meta.attempt += 1
-            if task.meta.max_retries > 0 and new_task.meta.attempt > new_task.meta.max_retries:
+            max_retries = min(task.meta.max_retries, self.settings.max_retries) \
+                if task.meta.max_retries else self.settings.max_retries
+
+            if new_task.meta.attempt > max_retries:
                 await self.backoff_task(task, reason="Max number of retries exceeded")
                 return
 
@@ -345,7 +370,7 @@ class BaseExecutor(Module):
             not_before = datetime.utcnow().timestamp() + countdown
 
         if not_before is not None:
-            new_task.meta.not_before.FromDatetime(not_before)
+            new_task.meta.not_before = not_before
 
         publisher = self.worker.get_publisher()
         await publisher.send_task(new_task)
@@ -392,18 +417,24 @@ class BaseExecutor(Module):
             for middleware in self._middlewares:
                 finalizer = (await middleware(task, receiver, finalizer)) or finalizer
 
-            finalizer = partial(finalizer, task=task, receiver=receiver)
-
             exec_fut = await self._execute(task, receiver)
         except Exception as exc:
             exec_fut = self.app_loop.create_future()
             exec_fut.set_exception(exc)
 
+        finalizer = partial(finalizer, task=task, receiver=receiver)
+
         exec_fut.add_done_callback(
             partial(self._task_done_callback, finalizer=finalizer(processing_fut=exec_fut), task_future=task_future)
         )
 
-    def _task_handling_completed(self, task_future):
+    def _task_handling_completed(self, task_future: asyncio.Future):
+        exc = task_future.exception()
+        if exc is not None:
+            # debug, there is many exceptions which are OK, if there was an error
+            # it should be already reported
+            self.logger.debug(f"Task future ended with exception {exc.__class__}({exc})")
+
         self._current_tasks.remove(task_future)
         if not self._current_tasks:
             self._busy.set(False)
@@ -415,14 +446,17 @@ class BaseExecutor(Module):
         :param receiver:
         :return:
         """
-        for router in self._routers:
+        for prefix, router in self._routers:
+            if not task.task_type.startswith(prefix):
+                continue
+
             processor = router.handle(task, receiver)
             if processor is not None:
                 # await for submit which may hold if processing unit is full
                 pu_fut = await self.unit.submit(processor)
                 return pu_fut
 
-        self.logger.error(f"No processor to handle message type '{task.task_type}'")
+        self.logger.error(f"No route to handle message type '{task.task_type}'")
         raise TaskNoHandler()
 
     def _task_done_callback(self,
@@ -465,11 +499,11 @@ class BaseExecutor(Module):
         handle_fut.add_done_callback(_callback)
 
     @lru_cache(maxsize=None)
-    def _get_exception_handler(self, exception) -> Iterator[ExceptionHandler]:
+    def _get_exception_handler(self, exc_class: Type[Exception]) -> Iterator[ExceptionHandler]:
         handlers = []
 
         for exc_type, callbacks in self._exc_handlers.items():
-            if not isinstance(exception, exc_type):
+            if not issubclass(exc_class, exc_type):
                 continue
 
             done = False
@@ -499,11 +533,17 @@ class BaseExecutor(Module):
         """
         exc = processing_fut.exception()
         if processing_fut.cancelled():
-            self.logger.info(f"CANCELLED: Task '{task.task_type} (uuid={task.uuid})' execution cancelled internally")
+            self.logger.info(f"Task '{task.task_type} (uuid={task.uuid})' execution cancelled internally")
             return
         elif exc is not None:
-            self.logger.warning(f"ERROR: Task '{task.task_type} (uuid={task.uuid})' execution failed with exception {exc.__class__.__name__}({exc})")
-            handlers = self._get_exception_handler(exc)
+            if not isinstance(exc, TaskControlException):
+                self.logger.warning(f"Task '{task.task_type} (uuid={task.uuid})' "
+                                    f"execution failed with exception {exc.__class__.__name__}({exc})")
+            else:
+                self.logger.info(f"Task '{task.task_type} (uuid={task.uuid})' "
+                                 f"ended with TCE {exc.__class__.__name__}")
+
+            handlers = self._get_exception_handler(exc.__class__)
             finalized = False
 
             for handler in handlers:
@@ -515,14 +555,16 @@ class BaseExecutor(Module):
                 try:
                     await handler(task, exc)
                 except TaskControlException as control_exc:
+                    self.logger.debug(f"Task '{task.task_type} (uuid={task.uuid})' "
+                                      f"exception handler '{handler}' requested TCE {exc.__class__.__name__}")
                     await self._handle_task_control(task, control_exc)
                     finalized = True
-                except Exception as exc:
-                    self.logger.exception(f"exception handler failed: {exc.__class__.__name__}({exc})")
+                except Exception as e:
+                    self.logger.exception(f"exception handler failed: {e.__class__.__name__}({e})")
                     if self.worker.debug:
-                        raise exc
+                        raise e
         else:
-            self.logger.info(f"DONE: Task '{task.task_type} (uuid={task.uuid})' processing completed")
+            self.logger.info(f"Task '{task.task_type} (uuid={task.uuid})' processing completed")
         await self._commit_task(task, receiver)
 
     async def _handle_task_control(self, task: Task, exc: Exception):
