@@ -2,13 +2,14 @@
 import asyncio
 
 import pytest as pytest
+import pytest_asyncio
 from patchwork.core import Task
 
 from patchwork.node import PatchworkWorker
 from patchwork.node.core.router import TaskRouter
 
 
-@pytest.fixture
+@pytest_asyncio.fixture
 async def worker():
     worker = PatchworkWorker(settings={
         'subscriber': {
@@ -25,6 +26,26 @@ async def worker():
     await worker.terminate()
 
 
+@pytest.mark.asyncio
+async def test_unsupported_task_use(worker):
+
+    caught_tasks = []
+
+    router = TaskRouter(skip_unsupported_tasks=True)
+    worker.include_router(router)
+
+    async def send():
+        task = Task(task_type='test-msg-2', payload=b'test')
+        task.meta.queue_name = 'test'
+        await worker.get_publisher().send_task(task)
+
+    asyncio.create_task(send())
+
+    # wait until executor processing unit becomes busy which means given message has been consumed
+    await asyncio.wait_for(worker.executor.busy.wait(), timeout=1)
+    assert len(caught_tasks) == 0
+
+
 @pytest.mark.asyncio
 async def test_simple_use(worker):
 
@@ -39,12 +60,12 @@ async def test_simple_use(worker):
     worker.include_router(router)
 
     async def send():
-        task = Task(task_type='test-msg')
+        task = Task(task_type='test-msg', payload=b'test')
         task.meta.queue_name = 'test'
         await worker.get_publisher().send_task(task)
 
     asyncio.create_task(send())
 
     # wait until executor processing unit becomes busy which means given message has been consumed
-    await asyncio.wait_for(worker.executor.unit.busy.wait(), timeout=1)
+    await asyncio.wait_for(worker.executor.busy.wait(), timeout=1)
     assert len(caught_tasks) == 1
